
from typing import Any, List, Callable


class Subscriber:
    """
    A class that subscribes to a Publisher and receives notifications when data changes.

    Args:
        callback (Callable[[Any], None]): A function that will be called whenever new data is published.

    Raises:
        TypeError: if callback is not a Callable class (lambda, function, etc)
    """

    def __init__(self, callback: Callable[[Any], None]):
        if not callable(callback):
            raise TypeError("Subscriber callback must be a callable function.")
        self.__callback = callback

    def _notify(self, data: Any):
        """
        Called when new data is published.

        Args:
            data (Any): The updated data from the publisher.
        """
        self.__callback(data)




from typing import Any, Callable

from ..datatype import SafeSet

from .Subscriber import Subscriber


class Publisher:
    """
    A thread-safe class that maintains a list of Subscriber instances and notifies them when data changes.    
    """

    def __init__(self):
        self.__subscribers = SafeSet()
        self.__data: Any = None

    def subscribe(self, subscriber: Subscriber):
        """
        Adds a subscriber to receive notifications.

        Args:
            subscriber (Subscriber): The subscriber instance.

        Raises:
            TypeError: if subscriber is not an instance of Subscriber class
        """
        if not isinstance(subscriber, Subscriber):
            raise TypeError("Expected an instance of Subscriber.")
        self.__subscribers.add(subscriber)

    def unsubscribe(self, subscriber: Subscriber):
        """
        Removes a subscriber from notifications.

        Args:
            subscriber (Subscriber): The subscriber instance.
        """
        self.__subscribers.remove(subscriber)

    def publish(self, data: Any):
        """
        Publishes new data and notifies all subscribers.

        Args:
            data (Any): The new data to be published.
        """
        self.__data = data
        # notify subscribers
        for subscriber in self.__subscribers:
            subscriber: Subscriber
            subscriber._notify(self.__data)




import queue

from typing import Any, Callable, Iterable, Self, Type

from .utils import *

from ..thread.ThreadBase import ThreadBase


class PipelineStage:
    """
    A pipeline stage that processes data through a callback function.
    It can run N separate threads to read and write data in parallel.

    This class allows data to be pushed to an input queue, where it is processed
    by the provided callback function, and the result is then placed in an output
    queue. This can be useful for concurrent processing of tasks in a pipeline
    fashion.

    The pipeline runs indefinetely, until .stop() is called.

    Args:

        callback (Callable): The function (or callable) that processes input data
                              and produces output. The callback should accept one
                              argument and return the processed result.

    Raises:

        ThreadBase.CallableException: If the provided callback is not callable.

        ValueError: If `n_threads` < 1.
    """

    EmptyException = queue.Empty
    """
    Raised when one of the following conditions happens:
    - get(block=False) is called, and there is no input in IN_QUEUE
    - get(timeout=value) and timeout exceeded (no input received within timeout time frame)
    """

    FullException = queue.Full
    """
    Raised when one of the following conditions happens:
    - put(block=False) is called and there is no available space in the OUT_QUEUE
    - put(timeout=value) and timeout exceeded (OUT_QUEUE full and timeout has expired)
    """

    StoppedException = queue.ShutDown
    """Raised when put()/get() is called after Pipeline.stop()"""

    @staticmethod
    def is_instance(obj: Any):
        """
        Checks if obj is an instance of PipelineStage

        Args:

            obj (Any): Object to check

        Raises:

            TypeError: if obj is not an instance of PipelineStage

        Returns:

            PipelineStage: object
        """
        if not isinstance(obj, PipelineStage):
            raise TypeError("Object is not a Pipeline Stage.")
        return obj

    def __init__(self, callback: Callable, n_threads: int = 1):
        """
        Initializes the pipeline stage with a callback function.

        Args:

            callback (Callable): The function to process data through the pipeline stage.
            n_threads (int): Number of threads that will read the input queue, and store result in output queue.

        Raises:

            ThreadBase.CallableException: If the callback argument is not callable.
            ValueError: If `n_threads` < 1.
        """

        self.__callback: Callable = ThreadBase.is_callable(callback)
        self.__input_queue = queue.Queue()
        self.__output_queue = queue.Queue()
        self.__started = False
        self.__threads: list[ThreadBase] = []

        if n_threads < 1:
            raise ValueError(
                "At least one thread is needed to run PipelineStage")

        for i in range(n_threads):
            self.__threads.append(
                ThreadBase(self.__run_pipeline, repeat=True)
            )

    def __run_pipeline(self):
        """
        Method to be executed in the thread. It gets data from the input queue,
        processes it through the callback function, and puts the result into
        the output queue.

        Raises:

            queue.Full: If the output queue is full (no available slot to store output)

        """
        try:
            input_data = self.__input_queue.get()
            output_data = self.__callback(input_data)
            self.__output_queue.put(output_data)
        except queue.ShutDown as e:
            self.stop()

    def has_started(self) -> bool:
        """
        Checks if the pipeline stage has started.

        Returns:

            bool: True if pipeline stage has started, otherwise False.
        """
        return self.__started

    def is_alive(self) -> bool:
        """
        Checks if the pipeline stage is alive.

        Returns:

            bool: True if any thread of pipeline stage is still alive, otherwise False.
        """
        result = False
        for thread in self.__threads:
            result = result or thread.is_alive()
        return result

    def is_terminated(self) -> bool:
        """
        Checks if the pipeline stage has terminated.

        Returns:

            bool: True if pipeline stage HAS started and is NOT alive, otherwise False.
        """
        return self.has_started() and not self.is_alive()

    def put(self, value, block: bool = True, timeout: float | None = None):
        """
        Puts data into the input queue for processing.

        Args:

            value (Any): The data to be processed by the pipeline.

            block (true): Block until data can be inserted in queue

            timeout (float, optional): Timeout for the put operation.

        Raises:

            FullException: if block = True and timeout is exceeded, or
                if block = False and there is no available space in the IN queue

            StoppedException: if pipeline has stopped
        """
        self.__input_queue.put(value, block, timeout)

    def get(self, block: bool = True, timeout: float | None = None):
        """
        Retrieves the processed data from the output queue.

        Args:

            block (true): Block until data can be get from queue

            timeout (float, optional): Timeout for the get operation.

        Returns:

            Any: The processed data after passing through the callback function.

        Raises:

            EmptyException: if block = True and timeout is exceeded, or
                if block = False and no output is available in the OUT queue
        """
        return self.__output_queue.get(block, timeout)

    def connect_output(self, other_pipeline: Self):
        """
        Connects this Pipeline Stage output to the input of other_pipeline

        Args:

            other_pipeline (Self): other pipeline stage
        """
        self.__output_queue = other_pipeline.__input_queue

    def start(self):
        """
        Starts the pipeline stage threads

        Raises:
            RuntimeError: if start() is called more than once on the same thread object.
        """
        for thread in self.__threads:
            thread.start()
        self.__started = True

    def stop(self):
        """
        Stops the pipeline thread (immediately)
        """
        # stops threads' main loops
        for thread in self.__threads:
            try_except_finally_wrap(lambda: thread.stop())
        # prevent in/out queues from storing data
        try_except_finally_wrap(
            lambda: self.__input_queue.shutdown(immediate=True)
        )
        try_except_finally_wrap(
            lambda: self.__output_queue.shutdown(immediate=True)
        )

    def join(self, timeout: float | None = None):
        """
        Joins the pipeline stages' threads, waiting for them to finish.

        Args:

            timeout (float, optional): The maximum time to wait for threads to finish. Defaults to None.

        Raises:

            RuntimeError: if an attempt is made to join the current thread (main thread), or the join() is called before start()
        """
        for thread in self.__threads:
            thread.join(timeout)

    def stop_join(self, timeout: float | None = None):
        """
        Calls stop() and join() to stop the PipelineStage and waiting for its threads to finish.

        Args:

            timeout (float, optional): The maximum time to wait for threads to finish. Defaults to None.

        Raises:

            RuntimeError: if an attempt is made to join the current thread (main thread), or the join() is called before start()
        """
        self.stop()
        self.join(timeout=timeout)




# safethread/utils/__init__.py

"""
This module provides threaded classes that inherit from ThreadBase.

Classes:
- **Scheduler**: A class that runs a scheduled Callable (function, lambda, etc), after a pre-defined timeout, either singleshot or periodically.
- **Subprocess**: A class that runs a subprocess within a separate thread.
- **ThreadBase**: An abstract class manages threads.
"""

from .Scheduler import Scheduler
from .Subprocess import Subprocess
from .ThreadBase import ThreadBase




# safethread/utils/__init__.py

"""
This module provides utility functions and classes.

Classes:
- **Factory**: A thread-safe class that provides a `create()` method to create objects dynamically based on certain parameters or configurations. This can be used for creating objects of various types at runtime, without tightly coupling the client code to specific class implementations.
- **Pipeline**: A thread-safe class that connects multiple ``thread.PipelineStage`` objects together (input_queue => Pipe 1 => Pipe 2 => ... => output_queue).
- **PipelineStage**: A thread-safe class that runs threads to processes data (using a Callable) from an Input Queue and places its output in an Output Queue.
- **Publisher**: A thread-safe class that maintains a list of Subscriber instances and notifies them when data changes.    
- **Singleton**: A thread-safe class that ensures a SINGLE INSTANCE of an object is created and shared throughout the application. This is useful for managing resources or configurations that need to be globally accessible and consistent across the system.    
- **Subscriber**: A thread-safe class that subscribes to a Publisher and receives notifications when data changes.
"""

from .Factory import Factory
from .Pipeline import Pipeline
from .PipelineStage import PipelineStage
from .Publisher import Publisher
from .Singleton import Singleton
from .Subscriber import Subscriber
from .utils import *




"""Utility functions for the library"""

from typing import Callable, Iterable


def try_except_finally_wrap(callback: Callable, callback_fail: Callable = lambda: None, callback_final: Callable = lambda: None):
    """
    Wraps callback in a try except finally block

    Args:
        callback (Callable): Callable to wrap in try block
        callback_fail (Callable): Callable to run if except block is executed
        callback_final (Callable): Callable to run in finally block
    """
    try:
        callback()
    except:
        callback_fail()
    finally:
        callback_final()
